import logging
import os
from multiprocessing import Process, Pool

from prometheus_client import multiprocess

from lemmings.worker import Worker
from lemmings.statistics import  Statistics
from lemmings.utils import _num, ProcessStat
from lemmings.utils.const import DONE, FAIL

class TestExecutor:
    def __init__(self, ansi_colors=False, debug=False):
        self.workers = []
        self.ansi_colors = ansi_colors
        self.debug = debug
        self.test_plans = 0
        self.available_plans = {}

    def add_test_plan(self, task_set, workers=1):
        for a in range(workers):
            self.workers.append(Worker(f"thread{a}", task_set, ansi_colors=self.ansi_colors, debug=self.debug))
        self.test_plans += 1
        return self

    def start_all(self, shared_state):
        """
        :param args: list of shared variables between workers[processes]
        """
        stat = Statistics()
        with Pool(len(self.workers)) as p:
            procs = [Process(target=self.start_process, args=[w, stat, shared_state]) for w in self.workers]
            for p in procs: p.start()
            for p in procs: p.join()
            pids = [x.pid for x in procs]
            for pid in pids:
                multiprocess.mark_process_dead(pid)
            self.done(stat)

    def start_process(self, worker, stat, shared_state):
        shared_state.init_multiprocess()
        worker.set_statistics(stat)
        worker.set_shared(shared_state)
        id, name = worker.id, worker.name
        task_set = worker.task_set.__class__.__name__
        worker.set_title(f"{task_set} [pid {os.getpid():6d}<:{os.getppid()}]")
        worker.log("start process")
        worker.gc = ProcessStat()
        worker.execute()

    def done(self, stat):
        logging.info(f"merged results: {_num(stat.shared_executed)} {DONE},"
              f" {_num(stat.shared_failed)} {FAIL} "
              f"[{_num(stat.shared_total)} total]")



import logging
import multiprocessing

class SharedValue(object):
    """A multi-process shared value with automatic lock"""
    def __init__(self, value=0, type="i"):
        """
        :param type: see full list of possible in standard module: ctypes.*
        """
        self._val = multiprocessing.Value(type, value)
        # self.name = "TEST"
        # self._debug(f"init with {value}")
        self.name = ""

    def with_name(self, name):
        self.name = name
        return self

    @property
    def lock(self):
        return self._val.get_lock()

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __cmp__(self, other):
        other = self._get(other)
        if not isinstance(other, type(self._val)):
            raise TypeError(other, type(self._val), type(other))
        with self.lock:
            if self._val.value == other:
                return 0
            elif self._val.value > other:
                return 1
            return -1

    def __iadd__(self, other):
        self.__add__(other)
        return self

    def __imod__(self, other):
        self.__mod__(other)
        return self

    def __isub__(self, other):
        self.__sub__(other)
        return self

    # TODO may be bug? is it + or += ?
    def __add__(self, other):
        with self.lock:
            other = self._get(other)
            self._debug(f"add {other}, before")
            if not isinstance(other, int):
                raise NotImplementedError()
            self._val.value += other
            self._debug(f"add {other}, after")
            return self._val.value

    def __mod__(self, other):
        with self.lock:
            other = self._get(other)
            if not isinstance(other, int):
                raise NotImplementedError()
            return self._val.value % other

    def __sub__(self, other):
        self.__add__(-other)

    @property
    def value(self):
        with self.lock:
            self._debug("getter")
            return self._val.value

    @value.setter
    def value(self, _value):
        with self.lock:
            self._debug("setter before")
            _value = self._get(_value)
            if not isinstance(_value, type(self._val)):
                raise TypeError(_value, type(self._val), type(_value))
            self._val.value = _value
            self._debug("setter after")

    def _debug(self, operation):
        if self.name:
            logging.debug(f"{self.name} > {operation} : {self._val.value}")

    def _get(self, v):
        if self._type(v):
            return v.value
        return v

    def _type(self, v):
        return isinstance(v, SharedValue)

    def __format__(self, format_spec):
        return self._val.value.__format__(format_spec)

    def __str__(self):
        return str(self.value)


import asyncio
import functools
import inspect
import linecache
import logging
import os
import random
import sys
import traceback
from contextlib import contextmanager

from lemmings.utils.timer import Timer

timer = Timer("timed_execution", module="", clazz="unknown", method="unknown")


def random_by_weight(list, weight_func):
    weight_total = sum(weight_func(i) for i in list)
    n = random.uniform(0, weight_total)
    for idx, item in enumerate(list):
        if n < weight_func(item):
            return item
        n = n - weight_func(item)
    return None


def timed_execution(f):
    if asyncio.iscoroutinefunction(f):
        @functools.wraps(f)
        async def wrap(*args, **kwargs):
            clazz = args[0].__class__.__name__ if is_method(f) else "unknown"
            with timer(module=f.__module__, clazz=clazz, method=f.__name__):
                return await f(*args, **kwargs)
    else:
        @functools.wraps(f)
        def wrap(*args, **kwargs):
            clazz = args[0].__class__.__name__ if is_method(f) else "unknown"
            with timer(module=f.__module__, clazz=clazz, method=f.__name__):
                return f(*args, **kwargs)

    return wrap


def ignore_exception(func):
    @functools.wraps(func)
    def function_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            logging.exception("ignored exception")

    return function_wrapper


def is_method(fn):
    try:
        return inspect.getfullargspec(fn)[0][0] == 'self'
    except:
        return False


@contextmanager
def debug(file=sys.stdout):
    """usage:
        with debug(filename) as file:
            file.print(....)
        it will catch and print exceptions for debug
    """
    try:
        yield file
    except BaseException as e:
        print(e)
        traceback.print_stack(file=file)
        raise


def exception_with_line(e, root_cause=True):
    original = e
    if root_cause:
        while e.__cause__ is not None:
            e = e.__cause__
    msg = f"'{e}'"
    lines = get_traceback(e)
    if len(lines) > 0:
        _, filename, lineno, line = lines[len(lines) - 1]
        return f'{msg} @ {filename}:{lineno} "{first_chars(line, 20)}"'
    return msg


def first_chars(s, i=100):
    return s.ljust(i)[:i].strip()


def get_traceback(e):
    lines = []
    tb = e.__traceback__
    while tb is not None:
        path = tb.tb_frame.f_code.co_filename
        file = os.path.basename(path)
        linecache.checkcache(path)
        line = linecache.getline(path, tb.tb_lineno, tb.tb_frame.f_globals)
        lines.append((path, file, tb.tb_lineno, line.strip()))
        tb = tb.tb_next
    return lines



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lemmings", 
    version="0.9.1",
    author="radiantgeek",
    author_email='',
    license='MIT',
    description="Async load testing framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/radiantgeek/lemmings",
    packages=setuptools.find_packages(exclude=['examples', 'tests']),
    install_requires=[
        "httpx>=0.11.1", 
	"psutil>=5.7.0",
        "prometheus_client>=0.7.1"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)


import asyncio
import functools
import math
import time

from prometheus_client import values
from prometheus_client.metrics import MetricWrapperBase

_BUFFER_SIZE = 512

default_timer = time.perf_counter

class QTimer(object):
    def __init__(self, callback):
        self._callback = callback

    def _new_timer(self):
        return self.__class__(self._callback)

    def __enter__(self):
        self._start = default_timer()

    def __exit__(self, typ, value, traceback):
        duration = max(default_timer() - self._start, 0)
        self._callback(duration)

    def __call__(self, f):
        if asyncio.iscoroutinefunction(f):
            @functools.wraps(f)
            async def wrapper_timer(*args, **kwargs):
                with self._new_timer():
                    return await f(*args, **kwargs)
            return wrapper_timer
        else:
            @functools.wraps(f)
            def wrapper_timer(*args, **kwargs):
                with self._new_timer():
                    return f(*args, **kwargs)
            return wrapper_timer

class QuantileSummary(MetricWrapperBase):
    _type = 'summary'
    _reserved_labelnames = ['quantile']

    def _metric_init(self):
        self._count = values.ValueClass(self._type, self._name, self._name + '_count', self._labelnames, self._labelvalues)
        self._sum = values.ValueClass(self._type, self._name, self._name + '_sum', self._labelnames, self._labelvalues)
        self._created = time.time()

    def observe(self, amount):
        self._count.inc(1)
        self._sum.inc(amount)

    def time(self):
        return QTimer(self.observe)

    def route_time(self):
        return QTimer(self.observe)

    def _child_samples(self):
        return (
            ('_count', {}, self._count.get()),
            ('_sum', {}, self._sum.get()),
            ('_created', {}, self._created))


class Estimator(object):
    """Estimator is not concurrency safe."""

    def __init__(self, *invariants):
        """ Attributes:
            invariants: A list of floating point doubles containing the target
                quantile value and allowed error.   [(0.5, 0.01), (0.99, 0.001)]
                are the default if none are provided, signifying that the median
                will be provided at a one percent error limit and the 99th
                percentile at the a 0.1 percent error limit.
        """
        if not invariants:
            self._invariants = _DEFAULT_INVARIANTS
        else:
            self._invariants = [_Quantile(q, e) for (q, e) in invariants]
        self._buffer = []
        self._head = None
        self._observations = 0
        self._sum = 0

    def observe(self, value):
        """Samples an observation's value.
        Args:
            value: A numeric value signifying the value to be sampled.
        """
        self._buffer.append(value)
        if len(self._buffer) == _BUFFER_SIZE:
            self._flush()

        self._observations += 1
        self._sum += value

    def query(self, rank):
        self._flush()

        current = self._head
        if not current:
            return 0

        mid_rank = math.floor(rank * self._observations)
        max_rank = mid_rank + math.floor(self._invariant(mid_rank, self._observations) / 2)

        rank = 0.0
        while current._successor:
            rank += current._rank
            if rank + current._successor._rank + current._successor._delta > max_rank:
                return current._value

            current = current._successor

        return current._value

    def _flush(self):
        """Purges the buffer and commits all pending values into the estimator."""
        self._buffer.sort()
        self._replace_batch()
        self._buffer = []
        self._compress()

    def _replace_batch(self):
        """Incorporates all pending values into the estimator."""
        if not self._head:
            self._head, self._buffer = self._record(self._buffer[0], 1, 0, None), self._buffer[1:]

        rank = 0.0
        current = self._head

        for b in self._buffer:
            if b < self._head._value:
                self._head = self._record(b, 1, 0, self._head)

            while current._successor and current._value < b:
                rank += current._rank
                current = current._successor

            if not current._successor:
                current._successor = self._record(b, 1, 0, None)

            current._successor = self._record(b, 1, self._invariant(rank, self._observations)-1, current._successor)

    def _record(self, value, rank, delta, successor):
        """Catalogs a sample."""
        return _Sample(value, rank, delta, successor)

    def _invariant(self, rank, n):
        """Computes the delta value for the sample."""
        minimum = n + 1

        for i in self._invariants:
            delta = i._delta(rank, n)
            if delta < minimum:
                minimum = delta

        return math.floor(minimum)

    def _compress(self):
        """Prunes the cataloged observations."""
        rank = 0.0
        current = self._head

        while current and current._successor:
            if current._rank + current._successor._rank + current._successor._delta <= self._invariant(rank, self._observations):
                removed = current._successor

                current._value = removed._value
                current._rank += removed._rank
                current._delta = removed._delta
                current._successor = removed._successor

            rank += current._rank
            current = current._successor


class _Quantile(object):
    """_Quantile is an internal representation of an estimation target invariant.
    Attributes:
        quantile: A floating point value for the requested quantile along the [0, 1] interval.
        inaccuracy: A floating point value for the allowed error for the estimate along the [0, 1] interval.
    """
    def __init__(self, quantile, inaccuracy):
        self._quantile = quantile
        self._inaccuracy = inaccuracy
        self._coefficient_i = (2.0 * inaccuracy) / (1.0 - quantile)
        self._coefficient_ii = 2.0 * inaccuracy / quantile

    """Computes the delta for the observation."""
    def _delta(self, rank, n):
        if rank <= math.floor((self._quantile * n)):
            return self._coefficient_i * (n - rank)

        return self._coefficient_ii * rank



class _Sample(object):
    def __init__(self, value, rank, delta, successor):
        self._value = value
        self._rank = rank
        self._delta = delta
        self._successor = successor

_DEFAULT_INVARIANTS = [_Quantile(0.50, 0.01), _Quantile(0.99, 0.001)]



from lemmings.executor import *
from lemmings.shared import *
from lemmings.statistics import *
from lemmings.task import *
from lemmings.task_set import *
from lemmings.task_set_run import *
from lemmings.worker import *

from lemmings.utils.influx import *
from lemmings.utils.prometheus import *
